from genai.schemas.tunes_params import (
    CreateTuneParams,
    DownloadAssetsParams,
    TunesListParams,
)
from genai.services.request_handler import RequestHandler
from genai.utils.errors import to_genai_error
from genai.utils.request_utils import sanitize_params


class TunesRouter:
    TUNES = "/v1/tunes"
    TUNE_METHODS = "/v1/tune_methods"

    def __init__(self, service_url: str, api_key: str) -> None:
        self.service_url = service_url.rstrip("/")
        self.key = api_key

    def list_tunes(self, params: TunesListParams = None):
        """List all tunes on the server.

        Args:
            params (TunesListParams, optional): Parameters for the tune list.

        Returns:
            Any: json from querying for tune list.
        """
        try:
            params = sanitize_params(params)
            endpoint = self.service_url + TunesRouter.TUNES
            return RequestHandler.get(endpoint, key=self.key, parameters=params)
        except Exception as e:
            raise to_genai_error(e)

    def get_tune(self, tune_id: str):
        """Get a given tune from the server.

        Args:
            tune_id (str): Id of the tune to be retrieved.

        Returns:
            Any: json from querying for tune retrieval.
        """
        try:
            endpoint = self.service_url + TunesRouter.TUNES + "/" + tune_id
            return RequestHandler.get(endpoint, key=self.key)
        except Exception as e:
            raise to_genai_error(e)

    def create_tune(self, params: CreateTuneParams):
        """Create a new tune on the server.

        Args:
            params (CreateTuneParams): Parameters for the tune creation.

        Returns:
            Any: json with info about the created tune.
        """

        try:
            params = sanitize_params(params)
            endpoint = self.service_url + TunesRouter.TUNES
            return RequestHandler.post(endpoint, key=self.key, options=params)
        except Exception as e:
            raise to_genai_error(e)

    def delete_tune(self, tune_id: str):
        """Delete a tune from the server.

        Args:
            tune_id (str): Id of the tune to be deleted.

        Returns:
            Any: json with info about the deleted tune.
        """
        try:
            endpoint = self.service_url + TunesRouter.TUNES + "/" + tune_id
            return RequestHandler.delete(endpoint, key=self.key)
        except Exception as e:
            raise to_genai_error(e)

    def get_tune_methods(self):
        """Get list of tune methods.

        Returns:
            Any: json with info about the available tune methods.
        """
        try:
            endpoint = self.service_url + self.TUNE_METHODS
            return RequestHandler.get(endpoint, key=self.key)
        except Exception as e:
            raise to_genai_error(e)

    def download_tune_assets(self, params: DownloadAssetsParams):
        """Download tune asset.

        Returns:
            Any: json with info about the downloaded tune asset.
        """
        try:
            endpoint = self.service_url + TunesRouter.TUNES + "/" + params.id + "/content/" + params.content
            return RequestHandler.get(endpoint, key=self.key)
        except Exception as e:
            raise to_genai_error(e)



import logging

from genai.utils.general import to_model_instance

try:
    from transformers import Agent
except ImportError:
    raise ImportError(
        "Could not import HuggingFace transformers: Please install ibm-generative-ai[huggingface] extension."
    )


from typing import List, Optional

from genai import Credentials, Model
from genai.schemas import GenerateParams

logger = logging.getLogger(__name__)


class IBMGenAIAgent(Agent):
    def __init__(
        self,
        credentials: Credentials = None,
        model: Optional[str] = None,
        params: Optional[GenerateParams] = None,
        chat_prompt_template: Optional[str] = None,
        run_prompt_template: Optional[str] = None,
        additional_tools: Optional[List[str]] = None,
    ):
        super().__init__(
            chat_prompt_template=chat_prompt_template,
            run_prompt_template=run_prompt_template,
            additional_tools=additional_tools,
        )
        self.credentials = credentials
        self.model = model
        self.params = params

    def generate_one(self, prompt, stop):
        return self._generate([prompt], stop)[0]

    def generate_many(self, prompts, stop):
        return self._generate(prompts, stop)

    def _generate(self, prompts: List[str], stop: Optional[List[str]] = None) -> List[str]:
        result: List[str] = []
        if len(prompts) == 0:
            return result

        params = to_model_instance(self.params, GenerateParams)
        params.stop_sequences = stop or params.stop_sequences
        model = Model(model=self.model, params=params, credentials=self.credentials)
        for response in model.generate(prompts=prompts):
            generated_text = response.generated_text

            logger.info("Output of GENAI call: {}".format(generated_text))
            result.append(generated_text)

        return result



import re
from warnings import warn


class Credentials:
    DEFAULT_API = "https://workbench-api.res.ibm.com"

    def __init__(
        self,
        api_key: str,
        api_endpoint: str = DEFAULT_API,
    ):
        """
        Instantiate the credentials object

        Args:
            api_key (str): The GENAI API Key
            api_endpoint (str, optional): GENAI API Endpoint. Defaults to DEFAULT_API.
        """
        if api_key is None:
            raise ValueError("api_key must be provided")
        self.api_key = api_key
        if api_endpoint is None:
            raise ValueError("api_endpoint must be provided")
        self.api_endpoint = api_endpoint.rstrip("/")
        self._remove_api_endpoint_version()

    def _remove_api_endpoint_version(self) -> None:
        [api, *version] = re.split(r"(/v\d+$)", self.api_endpoint, maxsplit=1)
        if version:
            warn(
                DeprecationWarning(
                    f"The 'api_endpoint' property should not contain any explicit API version"
                    f"(rename it from '{self.api_endpoint}' to just '{api}')"
                )
            )
            self.api_endpoint = api



from genai.exceptions.genai_exception import GenAiException
from genai.services import ServiceInterface


def _get_service(credentials, service):
    if credentials is None and service is None:
        raise GenAiException(ValueError("Credentials or ServiceInterface must be provided."))
    if credentials is not None and service is not None:
        raise GenAiException(ValueError("Provide exactly one of Credentials or ServiceInterface but not both."))
    if credentials is not None:
        service = ServiceInterface(service_url=credentials.api_endpoint, api_key=credentials.api_key)
    return service



from abc import ABC, abstractmethod

from genai.schemas import GenerateParams, GenerateResult, TokenizeResult, TokenParams


class CustomModel(ABC):
    @property
    def model_id(self):
        """Model ID
        This is the ID that you would use when defining the model you want to use in genai

        example: "google/flan-t5-base"

        Raises:
            NotImplementedError: If you do not implement this property.
        """
        raise NotImplementedError

    @abstractmethod
    def generate(self, input_text: str, params: GenerateParams) -> GenerateResult:
        """Generate a response from your llm using the provided input text and parameters

        Args:
            input_text (str): The input prompt text
            params (GenerateParams): The parameters that the user code wishes to be used

        Raises:
            NotImplementedError: If you do not implement this function.

        Returns:
            GenerateResult: The result to be sent back to the client
        """
        raise NotImplementedError

    @abstractmethod
    def tokenize(self, input_text: str, params: TokenParams) -> TokenizeResult:
        """Tokenize the input text with your model and return the output

        Args:
            input_text (str): The input prompt text
            params (TokenParams): The parametersthat the user code wishes to be used

        Raises:
            NotImplementedError: If you do not implement this function.

        Returns:
            TokenizeResult: The result to be sent back to the client
        """
        raise NotImplementedError



from itertools import product

from genai.schemas import GenerateParams


def grid_search_generate_params(params_space: dict[str, list]) -> list[GenerateParams]:
    """
    Generate all combinations of parameters from a dictionary of lists

    Args:
        params_space (dict[str, list]): dictionary of params as lists

    Returns:
        list of GenerateParams
    """
    params_values = list(product(*params_space.values()))
    params_list = []
    for values in params_values:
        params = GenerateParams(**dict(zip(params_space.keys(), values)))
        params_list.append(params)
    return params_list



from typing import Literal, Union

try:
    from pandas import DataFrame
except ImportError:
    raise ImportError("Could not import pandas: Please install ibm-generative-ai[pandas] extension.")

from genai.exceptions import GenAiException
from genai.prompt_pattern import PromptPattern
from genai.utils.extensions import register_promptpattern_accessor

__all__ = ["PandasExtension"]


@register_promptpattern_accessor("pandas")
class PandasExtension:
    def __init__(self, prompt_pattern: PromptPattern):
        self._obj = prompt_pattern

    def sub_from_dataframe(
        self,
        dataframe: DataFrame,
        col_to_var: Union[dict, Literal["infer"]] = "infer",
        headers: bool = True,
        strategy: Literal["random", "sequential", "sample"] = "sequential",
        start_index: int = 0,
        n: int = 1,  # prompt number to be returned - Optional
    ) -> Union["PromptPattern", list["PromptPattern"]]:
        """
        Substitutes variables in the prompt pattern with values from a dataframe.

        Args:
            dataframe (pandas.DataFrame): The dataframe to use for substitution.
            col_to_var (Union[dict, Literal["infer"]], Optional): "infer": String or dict of column name to list of
                variables. For example, a user can specify a mapping like col_to_var={"column": list_of_vars,. ..}
                where the list against each column is a list of variables.
                If "infer", the col_to_var mapping is inferred based on the template's variables' name.
                For example, assuming a variable "name" and a column "name" in the tabular data, "infer" is equivalent
                to col_to_var={"name":["name"]}. In its most general form, a user can specify a mapping like
                col_to_var={"column": list_of_vars,. ..} where the list against each column is a list of variables which
                can be substituted from that column. Defaults to "infer".
            headers (bool, Optional): Indicates if the dataframe contains a header row. Only used if inferring
                col_to_var. Defalts to True.
            strategy (Literal[sequential, sample, random], optional): Define how the template will be populated from
                the given json.
                Sequential  : sequentially subs the template's variables from the json, starting at a given start_index.
                Sample      : randomly samples rows and instantiate prompts. Without repetition of the data.
                Random      : for each prompt variable, substitute the value from the corresponding key at random
                Defaults to "sequential".
            start_index (int, opttional): The index to start substitution from. Only used with strategy
                sequential. State the first row to be used for the substitution in the template. Defaults to 0.
            n (int, Optional): The number of prompts to return.
                - n = 1: default, returns a PromptPattern
                - n > 1: returns a list of PromptPattern up to min(n, maximum producible prompts from file)
                - n = -1: goes through the entire file.
                If n is -1 and strategy is Random, a GenAiException is raised.

        Returns:
            PromptPattern | list[PromptPattern]: A prompt pattern or a list of prompt patterns.
        """

        if strategy == "random" and n <= 0:
            raise GenAiException("n must be at least 1")

        if len(dataframe) < 1:
            raise ValueError("Dataframe seems to be empty.")

        columns = dataframe.columns.tolist()
        data = dataframe.values.tolist()

        if col_to_var == "infer":
            col_to_var, columns = self._obj._tabular_infer_mode_helper(columns)
            data = data if headers else ([list(dataframe.columns)] + data)
        if strategy == "sample":
            start_index = 0

        self._obj.validate_start_index(strategy, start_index, data)

        return self._obj._sub_from_tabular_data(data, columns, col_to_var, start_index, n, strategy)

    def sub_all_from_dataframe(
        self,
        dataframe: DataFrame,
        headers: bool = True,
        col_to_var: Union[dict, Literal["infer"]] = "infer",
    ) -> list["PromptPattern"]:
        """
        Substitutes variables in the prompt pattern with values from a dataframe.

        Args:
            dataframe (pandas.DataFrame): The dataframe to use for substitution.
            headers (bool, Optional): Indicates if the dataframe contains a header row. Only used if inferring
                col_to_var. Defalts to True.
            col_to_var (Union[dict, Literal["infer"]], Optional): "infer": String or dict of column name to list of
                variables. For example, a user can specify a mapping like col_to_var={"column": list_of_vars,. ..}
                where the list against each column is a list of variables.
                If "infer", the col_to_var mapping is inferred based on the template's variables' name.
                For example, assuming a variable "name" and a column "name" in the tabular data, "infer" is equivalent
                to col_to_var={"name":["name"]}. In its most general form, a user can specify a mapping like
                col_to_var={"column": list_of_vars,. ..} where the list against each column is a list of variables which
                can be substituted from that column. Defaults to "infer".

        Returns:
            list[PromptPattern]: A list of prompt patterns.
        """
        start_index = 0
        n = -1
        return self.sub_from_dataframe(dataframe, col_to_var, headers, "sequential", start_index, n)



import asyncio
import os

from dotenv import load_dotenv

from genai.schemas.history_params import HistoryParams
from genai.services import ServiceInterface

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()

api_key = os.getenv("GENAI_KEY")
SERVICE_URL = "https://workbench-api.res.ibm.com"

service = ServiceInterface(service_url=SERVICE_URL, api_key=api_key)
model = "google/ul2"
inputs = ["Hello! How are you today?"]

params = HistoryParams(
    limit=8,
    offset=0,
    status="SUCCESS",
    origin="API",
)

print("\n------------- Example (Async History)-------------\n")


async def app() -> None:
    async_history = await service.async_history(params=params)
    async_history = async_history.json()
    print(async_history)


if __name__ == "__main__":
    asyncio.run(app())



from httpx import AsyncClient

from genai.exceptions import GenAiException

__all__ = ["ConnectionManager"]

from genai.utils.http_provider import HttpProvider
from genai.utils.http_utils import AsyncRateLimitTransport, AsyncRetryTransport


class ConnectionManager:
    MAX_RETRIES_GENERATE = 5
    TIMEOUT_GENERATE = 600
    MAX_RETRIES_TOKENIZE = 5
    MAX_REQ_PER_SECOND_TOKENIZE = 10
    TIMEOUT = 600

    async_generate_client: AsyncClient = None
    async_tokenize_client: AsyncClient = None

    @staticmethod
    def make_generate_client():
        """Function to make async httpx client for generate."""
        if ConnectionManager.async_generate_client is not None:
            raise GenAiException(ValueError("Can't have two active async_generate_clients"))

        ConnectionManager.async_generate_client = HttpProvider.get_async_client(
            transport=AsyncRetryTransport(
                retries=ConnectionManager.MAX_RETRIES_GENERATE, **HttpProvider.default_http_transport_options
            ),
            timeout=ConnectionManager.TIMEOUT_GENERATE,
        )

    @staticmethod
    def make_tokenize_client():
        """Function to make async httpx client for tokenize."""
        if ConnectionManager.async_tokenize_client is not None:
            raise GenAiException(ValueError("Can't have two active async_tokenize_clients"))

        ConnectionManager.async_tokenize_client = HttpProvider.get_async_client(
            transport=AsyncRateLimitTransport(
                default_max_rate=ConnectionManager.MAX_REQ_PER_SECOND_TOKENIZE,
                default_time_period=1,
                retries=ConnectionManager.MAX_RETRIES_TOKENIZE,
                **HttpProvider.default_http_transport_options,
            )
        )

    @staticmethod
    async def delete_generate_client():
        """Function to delete async httpx client for generate."""
        if ConnectionManager.async_generate_client is not None:
            await ConnectionManager.async_generate_client.aclose()
            ConnectionManager.async_generate_client = None

    @staticmethod
    async def delete_tokenize_client():
        """Function to delte async httpx client for tokenize."""
        if ConnectionManager.async_tokenize_client is not None:
            await ConnectionManager.async_tokenize_client.aclose()
            ConnectionManager.async_tokenize_client = None



def watsonx_payload(template, data=None, files=None):
    _dict = {"data": {}}

    try:
        _dict["id"] = template.watsonx.id

        if data:
            _dict["data"] = data

        if files:
            _dict["data"]["example_file_ids"] = files

        return _dict

    except Exception as e:
        print(e)



from genai.extensions.huggingface.agent import IBMGenAIAgent
from genai.extensions.huggingface.save_huggingface import HuggingFaceDatasetExtension

__all__ = ["HuggingFaceDatasetExtension", "IBMGenAIAgent"]



import os
import pathlib

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.prompt_pattern import PromptPattern
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)
PATH = pathlib.Path(__file__).parent.resolve()

print("\n------------- Example (String Replacement)-------------\n")

params = GenerateParams(
    decoding_method="sample",
    max_new_tokens=1000,
    min_new_tokens=1,
    stream=False,
    temperature=0.7,
    top_k=50,
    top_p=1,
    random_seed=2,
)

creds = Credentials(api_key, api_endpoint)
model = Model("google/flan-ul2", params=params, credentials=creds)


# load prompt pattern from file
pt = PromptPattern.from_file(str(PATH) + os.sep + "templates" + os.sep + "instruction.yaml")
print("\nGiven template:\n", pt)

json_path = str(PATH) + os.sep + "assets" + os.sep + "seed_tasks.jsonl"
mapping = {
    "instruction": ["instruction1", "instruction2", "instruction3", "instruction4", "instruction5"],
    "input": ["input1", "input2", "input3", "input4", "input5"],
    "output": ["output1", "output2", "output3", "output4", "output5"],
}

# Build multiple prompt template using all available data from json
list_of_prompts = pt.sub_all_from_json(json_path=json_path, key_to_var=mapping)

# Build 'n' multiple prompt template using available data from json
list_of_prompts = pt.sub_from_json(json_path=json_path, key_to_var=mapping, strategy="sample", n=4)

print("-----------------------")
print("generated prompt")
print(list_of_prompts)
print("-----------------------")


responses = model.generate_as_completed(list_of_prompts)
for response in responses:
    print(f"Generated text: {response.generated_text}")



import os
from urllib.parse import urlparse

import pytest
from dotenv import load_dotenv

from genai import Credentials

pytestmark = pytest.mark.integration

load_dotenv()

GENAI_API = os.getenv("GENAI_API")
GENAI_KEY = os.getenv("GENAI_KEY")


@pytest.fixture
def credentials(request):
    return Credentials(api_key=GENAI_KEY, api_endpoint=GENAI_API)


@pytest.fixture
def non_mocked_hosts(credentials) -> list:
    api_endpoint = urlparse(credentials.api_endpoint)
    return [api_endpoint.hostname]


def pytest_exception_interact(node, call, report):
    excinfo = call.excinfo
    if "script" in node.funcargs:
        excinfo.traceback = excinfo.traceback.cut(path=node.funcargs["script"])
    report.longrepr = node.repr_failure(excinfo)



class SimpleResponse:
    @staticmethod
    def generate(**kwargs):
        SimpleResponse._check_for_errors("generate", **kwargs)

        response = {}
        response["id"] = "1cf9f510-5549-4ea5-a909-2cf9219c1bb5"
        response["model_id"] = kwargs["model"]
        response["created_at"] = "2023-03-15T18:28:12.007Z"
        response["results"] = [
            {
                "input_text": input,
                "generated_text": input,
                "generated_token_count": len(input.split(" ")),
                "input_token_count": len(input.split(" ")),
                "stop_reason": "MAX_TOKENS",
            }
            for input in kwargs["inputs"]
        ]

        return response

    @staticmethod
    def generate_stream(**kwargs):
        SimpleResponse._check_for_errors("generate", **kwargs)

        responses = []
        for idx, token in enumerate("This is the output text.".split(" ")):
            response = {
                "id": "1cf9f510-5549-4ea5-a909-2cf9219c1bb5",
                "model_id": kwargs["model"],
                "created_at": "2023-03-15T18:28:12.007Z",
                "results": [
                    {
                        "generated_text": token,
                        "generated_token_count": idx + 1,
                        "stop_reason": "MAX_TOKENS",
                    }
                ],
            }
            responses.append(response)

        return responses

    @staticmethod
    def generate_response_array_async(**kwargs):
        arr = []
        for i in range(len(kwargs["inputs"])):
            response = {}
            response["id"] = "1cf9f510-5549-4ea5-a909-2cf9219c1bb5"
            response["model_id"] = kwargs["model"]
            response["created_at"] = "2023-03-15T18:28:12.007Z"
            response["results"] = [
                {
                    "input_text": kwargs["inputs"][i],
                    "generated_text": " are stronger.",
                    "generated_token_count": 3,
                    "input_token_count": len(kwargs["inputs"][i].split(" ")),
                    "stop_reason": "MAX_TOKENS",
                }
            ]
            arr.append(response)
        return arr

    @staticmethod
    def tokenize(**kwargs):
        response = {}
        response["model_id"] = kwargs["model"]
        response["created_at"] = "2023-03-15T18:28:12.007Z"

        if "params" in kwargs and kwargs["params"].return_tokens is True:
            response["results"] = [
                {
                    "input_text": kwargs["inputs"][i],
                    "token_count": len(kwargs["inputs"][i].split(" ")),
                    "tokens": kwargs["inputs"][i].split(" "),
                }
                for i in range(len(kwargs["inputs"]))
            ]
        else:
            response["results"] = [
                {
                    "input_text": kwargs["inputs"][i],
                    "token_count": len(kwargs["inputs"][i].split(" ")),
                }
                for i in range(len(kwargs["inputs"]))
            ]

        return response

    @staticmethod
    def tokenize_response_array_async(**kwargs):
        arr = []
        num_inputs, batch_size = len(kwargs["inputs"]), 5
        num_batches = num_inputs // batch_size + (num_inputs % batch_size > 0)
        for i in range(num_batches):
            response = {}
            response["model_id"] = kwargs["model"]
            response["created_at"] = "2023-03-15T18:28:12.007Z"

            if "params" in kwargs and kwargs["params"].return_tokens is True:
                response["results"] = [
                    {
                        "input_text": kwargs["inputs"][i],
                        "token_count": len(kwargs["inputs"][i].split(" ")),
                        "tokens": kwargs["inputs"][i].split(" "),
                    }
                    for i in range(i * batch_size, min((i + 1) * batch_size, num_inputs))
                ]
            else:
                response["results"] = [
                    {
                        "input_text": kwargs["inputs"][i],
                        "token_count": len(kwargs["inputs"][i].split(" ")),
                    }
                    for i in range(i * batch_size, min((i + 1) * batch_size, num_inputs))
                ]
            arr.append(response)
        return arr

    @staticmethod
    def history(**kwargs):
        response = {
            "results": [
                {
                    "id": "some-id",
                    "duration": 431,
                    "request": {
                        "inputs": ["Write a tagline for an alumni association: Together we"],
                        "model_id": "google/ul2",
                        "parameters": {},
                    },
                    "status": "SUCCESS",
                    "created_at": "2022-12-19T22:53:22.000Z",
                    "response": SimpleResponse.generate(model="google/ul2", inputs=["some input"]),
                }
            ],
            "totalCount": 62,
        }
        return response

    @staticmethod
    def terms_of_use(**kwargs):
        response = {
            "results": {
                "tou_accepted": True,
                "tou_accepted_at": "2022-12-12T18:51:53.000Z",
                "firstName": "Thomas",
                "lastName": "Watson",
                "data_usage_consent": True,
                "generate_default": None,
            }
        }
        return response

    @staticmethod
    def error(status_code, error, message, extensions=None):
        err = {"status_code": status_code, "error": error, "message": message}

        if extensions is not None:
            err["extensions"] = extensions

        return err

    @staticmethod
    def _check_for_errors(func, **kwargs):
        if func == "generate":
            if "model" not in kwargs:
                return SimpleResponse.error(
                    404,
                    "Not found",
                    "Model not found",
                    {"code": "NOT_FOUND", "state": {}},
                )

            if "inputs" not in kwargs:
                return SimpleResponse.error(
                    400,
                    "Bad Request",
                    " must have required property 'inputs'",
                    {
                        "code": "INVALID_INPUT",
                        "state": [
                            {
                                "instancePath": "",
                                "params": {"missingProperty": "inputs"},
                            }
                        ],
                    },
                )

    @staticmethod
    def prompt_template(template, name):
        response = {
            "results": {
                "id": "5XU9Zv6mrG6KIACN",
                "name": name,
                "value": template,
                "created_at": "2023-05-08T11:51:18.000Z",
            }
        }
        return response

    @staticmethod
    def files(**kwargs):
        response = {}
        results = {
            "id": "dumb-id",
            "bytes": "931652",
            "file_name": "file_to_tune.json",
            "purpose": "tune",
            "storage_provider_location": "us-east",
            "created_at": "2023-05-17T14:55:06.000Z",
            "file_formats": [
                {
                    "id": "1",
                    "name": "generation",
                },
            ],
        }

        if "multipart_form_data" in kwargs:
            response = {"results": results}
        elif "file_id" in kwargs:
            response = {"results": results}
        elif "params" in kwargs:
            response = {
                "results": [results, results],
                "totalCount": 2,
            }
        else:
            response = {
                "results": [results],
                "totalCount": 1,
            }
        return response

    @staticmethod
    def tunes(**kwargs):
        from genai.schemas.tunes_params import CreateTuneParams

        response = {}
        results = {
            "id": "google/flan-t5-xxl-mpt-Bok9gSo3-2023-04-11-18-00-57",
            "name": "Tune #2 google/flan-t5-xxl (3B)",
            "model_id": "google/google/flan-t5-xxl",
            "model_name": "google/flan-t5-xxl (3B)",
            "status": "COMPLETED",
            "task_id": "generation",
            "parameters": {"batch_size": 4, "num_epochs": 12},
            "created_at": "2023-04-11T18:00:57.000Z",
        }

        if "tune_id" in kwargs:
            response["results"] = results
        elif "params" in kwargs and type(kwargs["params"]) == CreateTuneParams:
            response["results"] = results
        else:
            response = {
                "results": [results],
                "totalCount": 1,
            }
        return response

    @staticmethod
    def get_tune(**kwargs):
        response = {
            "results": {
                "id": "google/flan-t5-xl-mpt-oP8G21Dj-2023-04-11-18-11-54",
                "name": "Tune #1 google/flan-t5-xl (3B)",
                "model_id": "google/google/flan-t5-xl",
                "method_id": "mpt",
                "model_name": "google/flan-t5-xl (3B)",
                "status_message": "",
                "task_id": "generation",
                "status": "COMPLETED",
                "parameters": {"batch_size": 4, "num_epochs": 12},
                "created_at": "2023-04-11T18:11:54.000Z",
                "validation_files": [],
                "training_files": [
                    {
                        "id": "60d54ad5-b9d7-4acb-99d6-870ff31c9222",
                        "file_name": "file.json",
                        "created_at": "2023-04-24 10:26:02+02",
                    }
                ],
                "evaluation_files": [],
                "datapoints": {
                    "loss": [
                        {
                            "data": {"epoch": 0, "value": 1.9922},
                            "timestamp": "2023-05-07T09:05:56.000Z",
                        }
                    ]
                },
            }
        }

        return response

    @staticmethod
    def models(**kwargs):
        return {
            "results": [
                {
                    "id": "google/flan-t5-xl",
                    "name": "flan-t5-xl (3B)",
                    "size": "3B",
                    "source_model_id": None,
                    "token_limit": 4096,
                },
                {
                    "id": "flan-t5-xl-mpt-XmHNkJWk-2023-07-18-17-00-34",
                    "name": "flan-t5-xxl (11B)",
                    "size": "11B",
                    "source_model_id": "google/flan-t5-xl",
                    "token_limit": 4096,
                },
            ]
        }

    @staticmethod
    def create_tune(**kwargs):
        model_id = kwargs["model_id"] if "model_id" in kwargs else "google/flan-t5-xxl"
        name = kwargs["name"] if "name" in kwargs else "Tune #2 google/flan-t5-xxl (3B)"
        response = {
            "results": {
                "id": "google/flan-t5-xxl-mpt-Bok9gSo3-2023-04-11-18-00-57",
                "name": name,
                "model_id": model_id,
                "model_name": "google/flan-t5-xxl (3B)",
                "status": "COMPLETED",
                "task_id": "generation",
                "parameters": {"batch_size": 4, "num_epochs": 12},
                "created_at": "2023-04-11T18:00:57.000Z",
            }
        }
        return response

    @staticmethod
    def get_tune_methods():
        response = {
            "results": [
                {"id": "pt", "name": "Prompt tuning"},
                {"id": "mpt", "name": "Multitask prompt tuning"},
            ]
        }
        return response



# A stripped down and modified version of some definitions in
# https://github.com/pandas-dev/pandas/blob/2.0.x/pandas/core/accessor.py
import logging

logger = logging.getLogger(__name__)


class AccessorRegistry:
    def __init__(self, key: str, accessor) -> None:
        self.key = key
        self.accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            return self.accessor
        accessor_obj = self.accessor(obj)
        object.__setattr__(obj, self.key, accessor_obj)
        return accessor_obj


def _register_accessor(key, cls):
    def wrapper(accessor):
        setattr(cls, key, AccessorRegistry(key, accessor))
        cls._accessors.add(key)
        return wrapper

    return wrapper


def register_promptpattern_accessor(name: str):
    from genai import PromptPattern

    # The line below defines a parameterized decorator
    # The effect of lines below is to call the wrapper inside the
    # paramterized decorator with accessor as "A".
    # @register_accessor("extension_name", PromptPattern)
    # class A():
    return _register_accessor(name, PromptPattern)



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)
creds = Credentials(api_key, api_endpoint)  # credentials object to access GENAI

# Instantiate parameters for text generation
params = GenerateParams(decoding_method="sample", max_new_tokens=10)

# Instantiate a model proxy object to send your requests
flan_ul2 = Model("google/flan-ul2", params=params, credentials=creds)

prompts = ["Hello! How are you?", "How's the weather?"]
for response in flan_ul2.generate_async(prompts):
    print(f"Prompt: {response.input_text}\nResponse: {response.generated_text}")



from typing import Optional

from pydantic import BaseModel, Extra, Field

# TODO: Update the descriptions import
from genai.schemas.descriptions import TunesAPIDescriptions as tx


class TunesListParams(BaseModel):
    """Class to hold the parameters for listing tunes."""

    class Config:
        anystr_strip_whitespace = True
        extra = Extra.forbid

    limit: Optional[int] = Field(None, description=tx.LIMIT, le=100)
    offset: Optional[int] = Field(None, description=tx.OFFSET)
    status: Optional[str] = Field(None, description=tx.STATUS)
    search: Optional[str] = Field(None, description=tx.SEARCH)


class CreateTuneHyperParams(BaseModel):
    """Class to hold the hyperparameters for creating tunes."""

    class Config:
        anystr_strip_whitespace = True
        extra = Extra.allow

    accumulate_steps: Optional[int] = Field(16, description=tx.ACCUMULATE_STEPS)
    batch_size: Optional[int] = Field(16, description=tx.BATCH_SIZE)
    learning_rate: Optional[float] = Field(0.3, description=tx.LEARNING_RATE)
    max_input_tokens: Optional[int] = Field(256, description=tx.MAX_INPUT_TOKENS)
    max_output_tokens: Optional[int] = Field(128, description=tx.MAX_OUTPUT_TOKENS)
    num_epochs: Optional[int] = Field(20, description=tx.NUM_EPOCHS)
    num_virtual_tokens: Optional[int] = Field(100, description=tx.NUM_VIRTUAL_TOKENS)
    verbalizer: Optional[str] = Field("Input: {{input}} Output:", description=tx.VERBALIZER)
    init_method: Optional[str] = Field(None, description=tx.INIT_METHOD)
    init_text: Optional[str] = Field(None, description=tx.INIT_TEXT)


class CreateTuneParams(BaseModel):
    """Class to hold the parameters for creating tunes."""

    class Config:
        anystr_strip_whitespace = True
        extra = Extra.allow

    name: str = Field(None, description=tx.NAME)
    model_id: str = Field(None, description=tx.MODEL_ID)
    method_id: str = Field(None, description=tx.METHOD_ID)
    task_id: str = Field(None, description=tx.TASK_ID)
    training_file_ids: list[str] = Field(None, description=tx.TRAINING_FILE_IDS)
    validation_file_ids: Optional[list[str]] = Field(None, description=tx.VALIDATION_FILE_IDS)
    parameters: Optional[CreateTuneHyperParams] = Field(None, description=tx.PARAMETERS)


class DownloadAssetsParams(BaseModel):
    """Class to hold the parameters for downloading tune assets."""

    class Config:
        anystr_strip_whitespace = True
        # extra: Extra.forbid

    id: str = Field(None, description=tx.ID)
    content: str = Field(None, description=tx.CONTENT)



import asyncio
import heapq
import logging
import math
import threading
from asyncio import Future
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from signal import SIGINT, SIGTERM, signal
from typing import Optional

from genai.exceptions import GenAiException
from genai.options import Options
from genai.schemas.responses import GenerateResponse, TokenizeResponse
from genai.services.connection_manager import ConnectionManager
from genai.services.service_interface import ServiceInterface
from genai.utils.errors import to_genai_error

logger = logging.getLogger(__name__)

__all__ = ["AsyncResponseGenerator"]


class AsyncResponseGenerator:
    LIMITS_CHECK_SLEEP_DURATION: float = 1.0

    def __init__(
        self,
        model_id,
        prompts,
        params,
        service: ServiceInterface,
        fn="generate",
        ordered=False,
        callback=None,
        options: Optional[Options] = None,
        *,
        throw_on_error: bool = False,
        max_concurrency_limit: Optional[int] = None,
    ):
        """Instantiates the ConcurrentWrapper Interface.

        Args:
            model_id (str): The type of model to use
            prompts (list): List of prompts
            params (GenerateParams): Parameters to use during generate requests
            service (ServiceInterface): The service interface
            fn (Literal["generate", "tokenize"]): Function to call from service
            callback (Callable[[GenerateResult], Any]): callback to execute once a prompt is successfully received.
        """
        self.model_id = model_id
        self.prompts = prompts
        self.params = params
        self.service = service
        self.callback = callback
        self.fn = fn
        self.ordered = ordered
        self.options = options
        self.throw_on_error = throw_on_error
        self._client_close_fn = None
        self._is_terminating = False
        self._max_concurrency_limit = max_concurrency_limit

    def __enter__(self):
        self._accumulator = []
        self._initialize_fn_specific_params()
        self._queue = Queue()
        self._loop = asyncio.new_event_loop()
        self._is_terminating = False
        return self

    def _shutdown(self):
        pending = asyncio.all_tasks(self._loop)
        for t in pending:
            t.cancel()
        if self._loop.is_running():
            self._loop.stop()
        while not self._queue.empty():
            _ = self._queue.get()
            self._queue.task_done()
        self._queue.join()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self._shutdown()
        except Exception as e:
            logger.error(str(e))
            raise to_genai_error(e)

    def _initialize_fn_specific_params(self):
        if self.fn == "generate":
            self._batch_size = 1
            self._num_batches = len(self.prompts)
            self._message_type = GenerateResponse
            self._service_fn = self.service.async_generate
            ConnectionManager.make_generate_client()
            self._client_close_fn = ConnectionManager.delete_generate_client
        elif self.fn == "tokenize":
            self._batch_size = 5
            a, b = divmod(len(self.prompts), self._batch_size)
            self._num_batches = a + (b > 0)
            self._message_type = TokenizeResponse
            self._service_fn = self.service.async_tokenize
            ConnectionManager.make_tokenize_client()
            self._client_close_fn = ConnectionManager.delete_tokenize_client

    def _generate_batch(self):
        for i in range(0, len(self.prompts), self._batch_size):
            yield self.prompts[i : min(i + self._batch_size, len(self.prompts))]

    def _process_response(self, response, batch_size):
        if response is None:
            for _ in range(batch_size):
                yield None
        else:
            for result in response.results:
                yield result

    async def _get_response_json(self, model, inputs, params, options):
        try:
            response_raw = await self._service_fn(model, inputs, params, options)
            response = response_raw.json()
            if response_raw and (200 < response_raw.status_code or response_raw.status_code > 299):
                raise Exception(response)
        except Exception as ex:
            logger.error("Error in _get_response_json {}: {}".format(type(ex), str(ex)))
            if self.throw_on_error:
                raise to_genai_error(ex)
            response = None
        return response

    async def _task(self, inputs, batch_num):
        response = None
        try:
            response = await self._get_response_json(self.model_id, inputs, self.params, self.options)
            logger.debug("Received response = {}".format(response))
            for i in range(len(response["results"])):
                response["results"][i]["input_text"] = inputs[i]
            response = self._message_type(**response)
            logger.debug("Cast to Response = {}".format(response))
        except Exception as e:
            logger.error(
                "Exception raised async_generate and casting : {}, response = {}, inputs = {}".format(
                    str(e), response, inputs
                )
            )
            self._queue.put_nowait((batch_num, len(inputs), None, to_genai_error(e)))
            return
        try:
            self._queue.put_nowait((batch_num, len(inputs), response, None))
            if self.callback is not None:
                for result in response.results:
                    self.callback(result)
        except Exception as e:
            logger.error(
                "Exception raised in callback : {}, response = {}, inputs = {}".format(str(e), response, inputs)
            )
            raise e

    async def _schedule_requests(self):
        local_concurrency_limit = max(self._max_concurrency_limit or math.inf, 1)

        async def get_limits():
            nonlocal local_concurrency_limit
            if local_concurrency_limit <= 0:
                return local_concurrency_limit

            if self.fn == "tokenize":
                return min(local_concurrency_limit, len(self.prompts))

            limits = self.service.generate_limits()
            tokens_remaining = limits.tokenCapacity - limits.tokensUsed
            return min(tokens_remaining, local_concurrency_limit)

        has_error = False

        def increment_local_concurrency_limit(future: Future):
            if future.exception():
                nonlocal has_error
                has_error = True

            nonlocal local_concurrency_limit
            local_concurrency_limit += 1

        tasks = []
        try:
            remaining_tokens = await get_limits()
            for idx, batch in enumerate(self._generate_batch()):
                while remaining_tokens <= 0 and not self._is_terminating and not has_error:
                    await asyncio.sleep(max(AsyncResponseGenerator.LIMITS_CHECK_SLEEP_DURATION, 0.0))
                    remaining_tokens = await get_limits()

                if has_error:
                    break

                if self._is_terminating:
                    raise GenAiException("Generation has been aborted by the user.")

                remaining_tokens -= 1
                local_concurrency_limit -= 1

                logger.debug("Creating task for batch_num {}".format(idx))
                task = asyncio.create_task(self._task(batch, idx))
                task.add_done_callback(increment_local_concurrency_limit)
                tasks.append(task)
        finally:
            await asyncio.gather(*tasks)

    def _request_launcher(self):
        try:
            asyncio.set_event_loop(self._loop)
            self._loop.run_until_complete(self._schedule_requests())
        except Exception as ex:
            self.throw_on_error = True
            self._queue.put_nowait((1, 1, None, ex))
            raise ex
        finally:
            self._loop.run_until_complete(self._cleanup())

    async def _cleanup(self):
        if self._client_close_fn is not None:
            await self._client_close_fn()

    def generate_response(
        self,
    ):  # -> Generator[Union[GenerateResult, TokenizeResult, None]]:
        """Method to spawn a launcher thread to launch requests for generate endpoint
        and yield responses as they get received.

        Yields:
            Generator[Union[GenerateResult, TokenizeResult, None]]: A generator of results
        """
        if len(self.prompts) == 0:
            return

        with ThreadPoolExecutor(max_workers=1) as executor:
            if threading.current_thread() is threading.main_thread():

                def init_termination(*args):
                    self._is_terminating = True
                    executor.shutdown(cancel_futures=False, wait=False)

                signal(SIGTERM, init_termination)
                signal(SIGINT, init_termination)

            counter = 0
            minheap, batch_tracker = [], 0

            task = executor.submit(self._request_launcher)
            while counter < self._num_batches:
                try:
                    batch_num, batch_size, response, error = self._queue.get()
                    self._queue.task_done()
                except Exception as ex:
                    logger.error("Exception while reading from queue: {}".format(str(ex)))
                    raise to_genai_error(ex)

                counter += 1

                if self.throw_on_error and error:
                    raise to_genai_error(error)

                try:
                    if not self.ordered:
                        for result in self._process_response(response, batch_size):
                            yield result
                    else:
                        heapq.heappush(minheap, (batch_num, batch_size, response, error))
                        for i in range(len(minheap)):
                            bnum, bsize, resp, error = heapq.heappop(minheap)
                            if bnum == batch_tracker:
                                for result in self._process_response(resp, bsize):
                                    yield result
                                batch_tracker += 1
                            else:
                                heapq.heappush(minheap, (bnum, bsize, resp, error))
                                break
                except Exception as ex:
                    logger.error("Error in processing: {}".format(str(ex)))
                    raise to_genai_error(ex)

            task_err = task.exception()
            if task_err is not None:
                raise task_err



import os
import pathlib

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.prompt_pattern import PromptPattern
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)
PATH = pathlib.Path(__file__).parent.resolve()

print("\n------------- Example (Greetings)-------------\n")

params = GenerateParams(
    decoding_method="sample",
    max_new_tokens=10,
    min_new_tokens=1,
    stream=False,
    temperature=0.7,
    top_k=50,
    top_p=1,
)

creds = Credentials(api_key, api_endpoint)
model = Model("google/flan-ul2", params=params, credentials=creds)

# can live locally or at an url endpoint
pt = PromptPattern.from_file(str(PATH) + os.sep + "templates" + os.sep + "capital-qa.yaml")
pt.sub("country", "Spain")
print(pt)
responses = model.generate_as_completed([str(pt)])
for response in responses:
    print(f"Generated text: {response.generated_text}")



import os
import pathlib
import shutil

import pytest


@pytest.mark.extension
class TestHuggingFace:
    def setup_method(self):
        self.path = pathlib.Path(__file__).parent.resolve()
        self.asset_path = pathlib.Path(__file__, "..", "..", "assets").resolve()

    def file_contains_string(self, prompt: str, path: str):
        with open(path + "/data-00000-of-00001.arrow", "rb") as f:
            if prompt.encode() in f.read():
                return True

    def test_huggingface(self, tmp_path):
        import genai.extensions.huggingface  # noqa
        from genai.prompt_pattern import PromptPattern

        datafile = pathlib.Path(self.asset_path, "age-gender.csv").resolve()

        # flake8: noqa
        template = """
            The age of the individual is {{0}}.
            The sex of the individual is {{1}}.
        """

        prompt = PromptPattern.from_str(template)

        list_of_prompts = prompt.sub_all_from_csv(csv_path=datafile, col_to_var="infer", headers=False)

        hfdataset = pathlib.Path(self.asset_path, "hf-dataset").resolve()

        prompt.huggingface.save_dataset(list_of_prompts, hfdataset)

        assert self.file_contains_string("The age of the individual is 31.", str(hfdataset))
        assert self.file_contains_string("The sex of the individual is female.", str(hfdataset))
        assert os.path.exists(hfdataset)

    def teardown_method(self):
        shutil.rmtree(pathlib.Path(self.asset_path, "hf-dataset").resolve())



import os
import pathlib
import random

from genai.prompt_pattern import PromptPattern

PATH = pathlib.Path(__file__).parent.resolve()

pt = PromptPattern.from_str(
    """
    Instruction: {{instruction}}
    Input: {{input}}
    Output: {{output}}
"""
)
print("\nGiven template:\n", pt)

json_path = str(PATH) + os.sep + "assets" + os.sep + "seed_tasks.json"

list_of_prompts = pt.sub_all_from_json(json_path=json_path, key_to_var="infer")

print("-----------------------")
print("Generated prompts: \n total number {}".format(len(list_of_prompts)))
print("Sample prompt: {}".format(list_of_prompts[random.randint(0, len(list_of_prompts) - 1)]))
print("-----------------------")



import re
import signal
from contextlib import nullcontext as does_not_raise
from unittest.mock import AsyncMock

import httpx
import pytest
from pytest_httpx import HTTPXMock

from genai import Credentials, Model
from genai.exceptions import GenAiException
from genai.schemas import GenerateParams, TokenParams
from genai.schemas.responses import GenerateResponse, GenerateResult, TokenizeResponse
from genai.services import ServiceInterface
from tests.assets.response_helper import SimpleResponse


@pytest.mark.unit
class TestModelAsync:
    def setup_method(self):
        self.service = ServiceInterface(service_url="SERVICE_URL", api_key="API_KEY")
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

    @pytest.fixture
    def mock_generate_json(self, mocker):
        async_mock = AsyncMock()
        mocker.patch(
            "genai.services.AsyncResponseGenerator._get_response_json",
            side_effect=async_mock,
        )
        return async_mock

    @pytest.fixture
    def generate_params(self):
        return GenerateParams(decoding_method="sample", max_new_tokens=5, min_new_tokens=0)

    @pytest.fixture
    def mock_tokenize_json(self, mocker):
        async_mock = AsyncMock()
        mocker.patch(
            "genai.services.AsyncResponseGenerator._get_response_json",
            side_effect=async_mock,
        )
        return async_mock

    @pytest.fixture
    def tokenize_params(self):
        return TokenParams(return_tokens=True)

    @pytest.mark.asyncio
    async def test_generate_async(self, mock_generate_json, generate_params):
        num_prompts = 31
        prompts = ["TEST_PROMPT"] * num_prompts
        expected = SimpleResponse.generate_response_array_async(
            model=self.model, inputs=prompts, params=generate_params
        )
        creds = Credentials("TEST_API_KEY")
        mock_generate_json.side_effect = expected

        model = Model(
            "google/flan-ul2",
            params=generate_params,
            credentials=creds,
        )

        counter = 0
        responses = list(model.generate_async(prompts, throw_on_error=True, hide_progressbar=True))
        assert len(responses) == len(expected)
        for batch_idx in range(len(expected)):
            for result in GenerateResponse(**expected[batch_idx]).results:
                assert responses[counter] == result
                assert prompts[counter] == result.input_text
                counter += 1
        assert counter == num_prompts

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "max_concurrency_limit,expectation,patch_generate_limits",
        [
            (1, does_not_raise(), {"tokens_capacity": 10}),
            (2, pytest.raises(AssertionError), {"tokens_capacity": 10}),
            (3, pytest.raises(AssertionError), {"tokens_capacity": 10}),
            (4, pytest.raises(AssertionError), {"tokens_capacity": 10}),
        ],
        indirect=["patch_generate_limits"],
    )
    async def test_generate_custom_concurrency_limit(
        self, generate_params, httpx_mock: HTTPXMock, max_concurrency_limit: int, expectation, patch_generate_limits
    ):
        generate_request_url = re.compile(f".+{ServiceInterface.GENERATE}$")

        num_prompts = 10
        prompts = ["TEST_PROMPT"] * num_prompts
        for prompt in prompts:
            response = SimpleResponse.generate(model=self.model, inputs=[prompt], params=generate_params)
            httpx_mock.add_response(url=generate_request_url, method="POST", json=response)

        model = Model(
            "google/flan-ul2",
            params=generate_params,
            credentials=Credentials("TEST_API_KEY"),
        )

        with expectation:
            for count, response in enumerate(
                model.generate_async(
                    prompts=prompts,
                    max_concurrency_limit=max_concurrency_limit,
                    throw_on_error=True,
                    hide_progressbar=True,
                    ordered=False,
                ),
                start=1,
            ):
                assert len(httpx_mock.get_requests(url=generate_request_url)) == count
                assert response is not None

    @pytest.mark.asyncio
    async def test_tokenize_async(self, mock_tokenize_json, tokenize_params):
        num_prompts = 31
        prompts = ["TEST_PROMPT " + str(i) for i in range(num_prompts)]
        expected = SimpleResponse.tokenize_response_array_async(
            model=self.model, inputs=prompts, params=tokenize_params
        )
        creds = Credentials("TEST_API_KEY")
        mock_tokenize_json.side_effect = expected

        model = Model("google/flan-ul2", params=tokenize_params, credentials=creds)

        counter = 0
        responses = list(model.tokenize_async(prompts))
        for batch_idx in range(len(expected)):
            for result in TokenizeResponse(**expected[batch_idx]).results:
                assert responses[counter] == result
                assert prompts[counter] == result.input_text
                counter += 1
        assert counter == num_prompts

    @pytest.mark.asyncio
    async def test_generate_callback(self, mock_generate_json, generate_params):
        num_prompts = 41
        single_response = SimpleResponse.generate(model=self.model, inputs=self.inputs, params=generate_params)
        creds = Credentials("TEST_API_KEY")
        mock_generate_json.return_value = single_response

        message = ""

        def tasks_completed(result):
            nonlocal message
            message += result.generated_text

        prompts = ["TEST_PROMPT"] * num_prompts
        model = Model("google/flan-ul2", params=generate_params, credentials=creds)

        for result in model.generate_async(prompts, callback=tasks_completed):
            pass

        assert message == GenerateResponse(**single_response).results[0].generated_text * num_prompts

    @pytest.mark.asyncio
    async def test_tokenize_callback(self, mock_tokenize_json, tokenize_params):
        num_prompts = 29
        inputs = ["Input input" for i in range(num_prompts)]
        expected = SimpleResponse.tokenize_response_array_async(model=self.model, inputs=inputs, params=tokenize_params)
        creds = Credentials("TEST_API_KEY")
        mock_tokenize_json.side_effect = expected

        message = []

        def tasks_completed(result):
            nonlocal message
            message += result.tokens

        prompts = ["TEST_PROMPT"] * num_prompts
        model = Model("google/flan-ul2", params=tokenize_params, credentials=creds)

        for result in model.tokenize_async(prompts, callback=tasks_completed):
            pass

        assert message == TokenizeResponse(**expected[0]).results[0].tokens * num_prompts

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "patch_generate_limits",
        [
            ({"tokens_capacity": 1}),
        ],
        indirect=["patch_generate_limits"],
    )
    async def test_generate_async_progress_bar(
        self, generate_params, patch_async_requests_limits, patch_generate_limits, httpx_mock: HTTPXMock
    ):
        """Tests that the response is yielded immediately (key part for the progress bar to work as expected)"""
        handled_responses = 0

        def handle_request(request: httpx.Request):
            nonlocal handled_responses
            handled_responses += 1

            return httpx.Response(json=SimpleResponse.generate(inputs=["Hello"], model=self.model), status_code=200)

        httpx_mock.add_callback(
            callback=handle_request,
            url=re.compile(f".+{ServiceInterface.GENERATE}$"),
            method="POST",
        )

        model = Model(self.model, params=generate_params, credentials=Credentials("TEST_API_KEY"))
        prompts = ["Hello world!"] * 10

        for count, response in enumerate(
            model.generate_async(prompts=prompts, hide_progressbar=True, throw_on_error=True),
            start=1,
        ):
            assert count == handled_responses
            assert response is not None

    @pytest.mark.asyncio
    @pytest.mark.parametrize("throw_in_callback", [True, False])
    @pytest.mark.parametrize("termination_signal", [signal.SIGINT, signal.SIGTERM])
    @pytest.mark.parametrize("execution_params", [{"throw_on_error": True}, {"throw_on_error": False}])
    async def test_generate_termination_results(
        self,
        generate_params,
        httpx_mock: HTTPXMock,
        execution_params: dict,
        throw_in_callback: bool,
        termination_signal: int,
    ):
        """Tests that the processing in case of crash ends up as soon as possible"""
        url = re.compile(f".+{ServiceInterface.GENERATE}$")
        httpx_mock.add_response(
            url=url,
            method="POST",
            json=SimpleResponse.generate(model=self.model, inputs=self.inputs),
        )

        model = Model(self.model, params=generate_params, credentials=Credentials("TEST_API_KEY"))
        prompts = ["Hello world!"] * 10
        chunk_size = 5
        has_raised = False

        def callback(result: GenerateResult):
            assert result is not None
            nonlocal has_raised
            if throw_in_callback and not has_raised:
                has_raised = True
                raise GenAiException("Unexpected Callback Error")

        with pytest.raises(GenAiException):
            for response in enumerate(
                model.generate_async(
                    **execution_params,
                    prompts=prompts,
                    hide_progressbar=True,
                    max_concurrency_limit=chunk_size,
                    callback=callback,
                ),
            ):
                assert response is not None
                # Abort as soon as possible
                if not has_raised and not throw_in_callback:
                    has_raised = True
                    signal.raise_signal(termination_signal)

        # Only one chunk is completed and then the process is terminated
        requests = httpx_mock.get_requests(url=url, method="POST")
        assert len(requests) == chunk_size



import logging

import pytest

from genai.model import Model
from genai.schemas import GenerateParams, GenerateResult, TokenizeResult, TokenParams

logger = logging.getLogger(__name__)


# Test implmentation of CustomModel class


@pytest.mark.extension
class TestLocalServer:
    def test_local_server(self):
        from genai.extensions.localserver import CustomModel, LocalLLMServer

        class ExampleModel(CustomModel):
            model_id = "example/model"

            def __init__(self):
                logger.info("Initialising my custom model")

            def generate(self, input_text: str, params: GenerateParams) -> GenerateResult:
                logger.info(f"Calling generate on: {input_text}")

                genai_response = GenerateResult(
                    generated_text=input_text,
                    generated_token_count=len(input_text.split(" ")),
                    input_token_count=len(input_text.split(" ")),
                    stop_reason="",
                )
                logger.info(f"Response to {input_text} was: {genai_response}")

                return genai_response

            def tokenize(self, input_text: str, params: TokenParams) -> TokenizeResult:
                logger.info(f"Calling tokenize on: {input_text}")
                tokens = input_text.split(" ")
                result = TokenizeResult(token_count=len(tokens))
                if params.return_tokens is True:
                    result.tokens = tokens
                return result

        # Instansiate the Local Server with your model
        server = LocalLLMServer(models=[ExampleModel])
        # Start the server and execute your code
        with server.run_locally():
            logger.info("Local Server Started, attempting basic generate call")
            # Get the credentials / connection details for the local server
            creds = server.get_credentials()

            # Instantiate parameters for text generation
            params = GenerateParams()

            # Instantiate a model proxy object to send your requests
            custom_model = Model(ExampleModel.model_id, params=params, credentials=creds)

            test_prompt = "hello this is a test of a custom model in a local server"
            responses = custom_model.generate([test_prompt])
            assert len(responses) == 1
            response = responses[0]
            # Our test model simply returns the input test, so this will verify that our server is working
            assert response.generated_text == test_prompt

            token_responses = custom_model.tokenize([test_prompt])
            assert len(token_responses) == 1
            token_response = token_responses[0]
            # Our test model simply returns the input test, so this will verify that our server is working
            assert token_response.token_count == len(test_prompt.split(" "))



import os
import time

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)

print("\n------------- Example (Model QA)-------------\n")

max_cycles = 20

bob_params = GenerateParams(
    decoding_method="sample",
    max_new_tokens=25,
    min_new_tokens=1,
    stream=False,
    temperature=1,
    top_k=50,
    top_p=1,
)

alice_params = GenerateParams(
    decoding_method="sample",
    max_new_tokens=45,
    min_new_tokens=1,
    stream=False,
    temperature=0.05,
    top_k=50,
    top_p=1,
)

creds = Credentials(api_key, api_endpoint)
bob_model = Model("google/flan-ul2", params=bob_params, credentials=creds)
alice_model = Model("google/flan-t5-xxl", params=alice_params, credentials=creds)

alice_q = "What is 1 + 1?"
print(f"[Alice][Q] {alice_q}")

for x in range(max_cycles):
    bob_response = bob_model.generate([alice_q])
    bob_a = bob_response[0].generated_text
    print(f"[Bob][A] {bob_a}")

    bob_q = "What is " + bob_a + " + " + bob_a + "?"
    print(f"[Bob][Q] {bob_q}")

    alice_response = alice_model.generate([bob_q])
    alice_a = alice_response[0].generated_text
    print(f"[Alice][A] {alice_a}")

    alice_q = "What is " + alice_a + " + " + alice_a + "?"
    print(f"[Alice][Q] {alice_q}")
    time.sleep(0.5)
